from __future__ import annotations
from abc import abstractmethod
import typing as t
from fsspec.spec import AbstractFileSystem as AbstractFsSpecFileSystem

from spice_rack import _bases, _logging
from spice_rack._fs_ops import _path_strs, _file_info, _open_modes, _exceptions


__all__ = (
    "AbstractFileSystem",
)


class AbstractFileSystem(
    _bases.dispatchable.DispatchableValueModelBase,
    _logging.log_extra.LoggableObjMixin
):
    """base class for all fsspec file system wrappers"""

    @abstractmethod
    def build_fsspec_file_system(self) -> AbstractFsSpecFileSystem:
        ...

    @property
    def fsspec_obj(self) -> AbstractFsSpecFileSystem:
        return self.build_fsspec_file_system()

    @classmethod
    @abstractmethod
    def get_fs_specific_prefix(cls) -> str:
        ...

    def clean_raw_path_str(
            self,
            raw_path: str,
    ) -> _path_strs.FileOrDirAbsPathT:
        assert isinstance(raw_path, str), type(raw_path)
        cleaned_path = raw_path.replace(
            self.get_fs_specific_prefix(),
            "/",
            1
        )
        return _path_strs.FileOrDirAbsPathTypeAdapter.validate_python(
            cleaned_path
        )

    def contextualize_abs_path(self, path: _path_strs.FileOrDirAbsPathT) -> str:
        """
        take an absolute path, in the general representation
        and add the file system specific prefix
        """
        if not isinstance(path, (_path_strs.AbsoluteFilePathStr, _path_strs.AbsoluteDirPathStr)):
            raise ValueError(f"should be an absolute path type, encountered type: {type(path)}")
        prefix_val = self.get_fs_specific_prefix()
        formatted_path_str = prefix_val + str(path)[1:]
        return formatted_path_str

    @abstractmethod
    def get_home_dir(self) -> _path_strs.AbsoluteDirPathStr:
        ...

    def __hash__(self) -> int:
        return hash(
            self.contextualize_abs_path(self.get_home_dir())
        )

    def __eq__(self, other: AbstractFileSystem) -> bool:
        if type(self) == type(other):  # noqa
            if hash(self) == hash(other):
                return True
        return False

    def special_repr(self) -> str:
        return f"FileSystem[{self.class_id}]"

    def __get_logger_data__(self) -> t.Dict:
        return {
            "file_system_type": self.class_id,
            "file_system_home_dir": self.contextualize_abs_path(self.get_home_dir())
        }

    @t.final
    def exists(self, path: _path_strs.FileOrDirAbsPathT) -> bool:
        """
        returns True if this file system object exists, false otherwise
        """

        # todo: what if perms issue not existence issue?
        return self.fsspec_obj.exists(self.contextualize_abs_path(path))

    def ensure_exists(self, path: _path_strs.FileOrDirAbsPathT) -> None:
        """
        ensure the file or dir path exists, raising an exception if not

        Args:
            path: the path to the file or directory

        Returns: Nothing

        Raises:
            NonExistentPathException: if the path doesn't exist, or we cannot access it

        """
        exists = self.exists(path)
        if not exists:
            raise _exceptions.NonExistentPathException(
                file_system=self,
                path=path,
            )
        else:
            return

    @staticmethod
    def ensure_correct_file_ext(
            path: _path_strs.RelOrAbsFilePathT,
            choices: list[_file_info.FileExt]
    ) -> None:
        """
        ensure the file path has one of the specified extensions
        Args:
            path: the path we are checking
            choices: the valid extensions

        Returns: Nothing

        Raises:
            FilePathInvalidException: if the file path has no extension of it isn't
                one of the choices

        """
        file_ext = path.get_file_ext()
        if file_ext is None:
            raise ValueError(
                f"'{path}' has no file extensions"
            )
            # raise _exceptions.FilePathInvalidException(
            #     file_system=self,
            #     path=path,
            #     reason="no file extension found on the file path",
            #     extra_info={
            #         "choices": choices,
            #     }
            # )

        if file_ext not in choices:
            raise ValueError(
                f"'{path}' file extension isn't one of {choices}"
            )
            # raise _exceptions.FilePathInvalidException(
            #     file_system=self,
            #     path=path,
            #     reason=f"the file extension, '{file_ext}', not one of the choices",
            #     extra_info={
            #         "file_path": path,
            #         "choices": choices,
            #     }
            # )
        return

    # def ensure_correct_mime_type(
    #         self,
    #         path: _path_strs.AbsoluteFilePathStr,
    #         choices: list[_special_types.MimeType]
    # ) -> None:
    #     """
    #     ensure the file path is one of the specified mime types
    #     Args:
    #         path: the path we are checking
    #         choices: the valid mime types
    #
    #     Returns: Nothing
    #
    #     Raises:
    #         FilePathInvalidException: if we cannot determine the mime type, or it is not one
    #             of the specified choices
    #     """
    #     mime_type = path.get_file_ext()
    #     if mime_type is None:
    #         raise _exceptions.FilePathInvalidException(
    #             file_system=self,
    #             path=path,
    #             reason="unable to infer mime type from the file path",
    #             extra_info={
    #                 "choices": choices,
    #             }
    #         )
    #     if mime_type not in choices:
    #         raise _exceptions.FilePathInvalidException(
    #             file_system=self,
    #             path=path,
    #             reason=f"the inferred mime type, '{mime_type}', not one of the choices",
    #             extra_info={
    #                 "choices": choices,
    #             }
    #         )
    #     return

    def open_file(
            self,
            path: _path_strs.AbsoluteFilePathStr,
            mode: _open_modes.SupportedOpenModesT
    ) -> _open_modes.OpenFileT:
        """
        return a readable open file object. todo: revisit type annotations of the return type here

        Args:
            path: the file path str
            mode: the mode we are opening in

        Returns:
            the readable open file object
        """
        if mode != "wb":
            self.ensure_exists(path=path)
        return self.fsspec_obj.open(path=self.contextualize_abs_path(path), mode=mode)

    def delete_file(
            self,
            path: _path_strs.AbsoluteFilePathStr,
            if_non_existent: t.Literal["raise", "return"] = "return"
    ) -> None:
        """
        delete the file
        """

        exists = self.exists(path)
        if not exists:
            if if_non_existent == "raise":
                self.ensure_exists(path)

        else:
            # what if perms issue not existence issue?
            self.fsspec_obj.delete(path=self.contextualize_abs_path(path), recursive=True)
        return

    def delete_dir(
            self,
            path: _path_strs.AbsoluteDirPathStr,
            recursive: bool = True,
            if_non_existent: t.Literal["raise", "return"] = "return"
    ) -> None:
        """
        delete the directory
        """

        from devtools import debug
        debug(path)

        exists = self.exists(path)

        debug(path)

        if not exists:
            if if_non_existent == "raise":
                self.ensure_exists(path)

        else:
            # what if perms issue not existence issue?
            self.fsspec_obj.delete(path=self.contextualize_abs_path(path), recursive=recursive)
        return

    def iter_dir_contents(
            self,
            path: _path_strs.AbsoluteDirPathStr,
    ) -> t.Iterator[_path_strs.FileOrDirAbsPathT]:
        """iterate over the top level dir contents"""
        for raw_info_rec_i in self.fsspec_obj.listdir(
            self.contextualize_abs_path(path)
        ):
            raw_path_i = raw_info_rec_i.get("name")
            if raw_path_i is None:
                raise ValueError(
                    "'name' not found in record?"
                )
                # raise _exceptions.FileSystemException(
                #     file_system=self,
                #     extra_info={
                #         "dir_path": path,
                #         "raw_record": raw_info_rec_i,
                #         "issue": "'name' not found in record?"
                #     },
                # )

            path_i: _path_strs.FileOrDirAbsPathT
            raw_type = raw_info_rec_i.get("type")
            formatted_path_i = str(self.clean_raw_path_str(raw_path_i))

            if raw_type == "file":
                path_i = _path_strs.AbsoluteFilePathStr(formatted_path_i)

            elif raw_type == "directory":
                if not formatted_path_i.endswith("/"):
                    formatted_path_i = f"{formatted_path_i}/"
                path_i = _path_strs.AbsoluteDirPathStr(formatted_path_i)

            else:
                raise ValueError(f"unexpected type val: '{raw_type}'")

            yield path_i

    def list_dir_contents(
            self,
            path: _path_strs.AbsoluteDirPathStr
    ) -> list[_path_strs.FileOrDirAbsPathT]:
        return list(self.iter_dir_contents(path=path))

    def iter_dir_contents_files_only(
            self,
            path: _path_strs.AbsoluteDirPathStr,
            recursive: bool = True
    ) -> t.Iterator[_path_strs.AbsoluteFilePathStr]:
        for path_i in self.iter_dir_contents(path=path):
            if isinstance(path_i, _path_strs.AbsoluteFilePathStr):
                yield path_i
            elif isinstance(path_i, _path_strs.AbsoluteDirPathStr):
                if recursive:
                    for path_ij in self.iter_dir_contents_files_only(
                        path=path_i,
                        recursive=recursive,
                    ):
                        yield path_ij
                else:
                    continue
            else:
                raise ValueError(type(path_i))

    def make_dir(
            self,
            path: _path_strs.AbsoluteDirPathStr,
            if_exists: t.Literal["raise", "return"] = "return",
            create_parents: bool = True
    ) -> None:
        if self.exists(path=path):
            if if_exists == "raise":
                raise ValueError(
                    f"'{path}'already exists, cannot make it."
                )
                # raise _exceptions.PathAlreadyExistsException(
                #     file_system=self,
                #     path=path,
                #     extra_info={
                #         "action_attempted": "make_dir"
                #     }
                # )
            else:
                return

        else:
            self.fsspec_obj.mkdir(
                path=self.contextualize_abs_path(path),
                create_parents=create_parents
            )

    def download_file_locally(
            self,
            path: _path_strs.AbsoluteFilePathStr,
            dest_dir: _path_strs.AbsoluteDirPathStr
    ) -> _path_strs.AbsoluteFilePathStr:
        from spice_rack._fs_ops._file_systems import _local
        local_fs = _local.LocalFileSystem()
        local_fs.make_dir(dest_dir, if_exists="return", create_parents=True)

        local_path = dest_dir.joinpath(
            _path_strs.RelFilePathStr(path.get_name(include_suffixes=True))
        )
        self.fsspec_obj.download(
            lpath=local_fs.contextualize_abs_path(local_path),
            rpath=self.contextualize_abs_path(path)
        )
        return local_path

    def download_dir_locally(
            self,
            path: _path_strs.AbsoluteDirPathStr,
            dest_dir: _path_strs.AbsoluteDirPathStr
    ) -> _path_strs.AbsoluteDirPathStr:
        from spice_rack._fs_ops._file_systems import _local
        from spice_rack._fs_ops._helpers import is_placeholder_file_path
        local_fs = _local.LocalFileSystem()
        local_fs.make_dir(dest_dir, if_exists="return", create_parents=True)

        local_path = dest_dir.joinpath(
            _path_strs.RelDirPathStr(path.get_name())
        )
        self.fsspec_obj.download(
            lpath=local_fs.contextualize_abs_path(local_path),
            rpath=self.contextualize_abs_path(path),
            recursive=True
        )

        # remove possible placeholder files
        for local_path_i in local_fs.iter_dir_contents_files_only(local_path):
            if is_placeholder_file_path(local_path_i):
                local_fs.delete_file(
                    local_path_i, if_non_existent="return"
                )
        return local_path



from spice_rack._fs_ops._fs_models import *
from spice_rack._fs_ops._file_info import *
from spice_rack._fs_ops import (
    _file_systems as file_systems,
    _path_strs as path_strs,
    _exceptions as exceptions,
)



from __future__ import annotations
from abc import abstractmethod
from devtools import pformat
import typing as t
import pydantic

from spice_rack import _bases, _logging
from spice_rack._fs_ops import (
    _path_strs,
    # _exceptions,
    _file_systems,
    _helpers,
    _open_modes,
    _file_info
)

__all__ = (
    "FilePath", "DirPath",
    "FileOrDirPathT", "FileOrDirPathTypeAdapter"
)


class _AbstractFileSystemObj(
    _bases.dispatchable.DispatchableValueModelBase,
    _logging.log_extra.LoggableObjMixin
):
    """
    Base class for all file system objects in this package
    All file and dir objects inherit this class
    """
    path: _path_strs.FileOrDirAbsPathT
    file_system: _file_systems.AnyFileSystemT = pydantic.Field(
        description="the file system object for this file",
        default=None
    )

    @pydantic.model_validator(mode="before")
    def _model_setup(cls, data: t.Any) -> t.Any:
        if isinstance(data, str):
            data = cls.init_from_str(data).model_dump()

        file_system_key = "file_system"
        path_key = "path"

        if isinstance(data, dict):
            raw_file_system = data.get(file_system_key)
            raw_path: str = data.get(path_key)
            if raw_path is None:
                raise ValueError(
                    f"no 'path' specified? data: \n{pformat(data)}"
                )
            else:
                raw_path = str(raw_path)
            file_system: _file_systems.AnyFileSystemT

            if raw_file_system is None:
                file_system = _helpers.infer_file_system(raw_path)
            else:
                file_system = _file_systems.AnyFileSystemTypeAdapter.validate_python(
                    raw_file_system
                )
            data["file_system"] = file_system

            # either way we want to clean the path
            # either way we clean the path
            path_formatted = file_system.clean_raw_path_str(raw_path)
            data[file_system_key] = file_system
            data[path_key] = path_formatted

        return data

    @property
    def file_system_type(self) -> _bases.dispatchable.ClassId:
        return self.file_system.class_id

    @t.final
    def exists(self) -> bool:
        """
        returns True if this file system object exists, false otherwise
        """
        return self.file_system.exists(self.path)

    def __repr__(self) -> str:
        """
        deprecated, we want to remove all usage of this
        in logging and error messaging and use special_repr only
        """
        return self.special_repr()

    def __hash__(self) -> int:
        return hash(
            (self.path, self.file_system)
        )

    @abstractmethod
    def _special_repr_short(self) -> str:
        ...

    @abstractmethod
    def _special_repr_verbose(self) -> str:
        ...

    @t.final
    def special_repr(self, verbose: bool = False) -> str:
        if verbose:
            return self._special_repr_verbose()
        else:
            return self._special_repr_short()

    def __eq__(self, other: _AbstractFileSystemObj) -> bool:
        if type(self) == type(other):  # noqa
            if hash(self) == hash(other):
                return True
        return False

    @abstractmethod
    def delete(
            self,
            if_non_existent: t.Literal["raise", "return"] = "return"
    ) -> None:
        ...

    def build_file_like(
            self,
            path: _path_strs.AbsoluteFilePathStr
    ) -> FilePath:
        return FilePath(
            path=path, file_system=self.file_system
        )

    def build_dir_like(
            self,
            path: _path_strs.AbsoluteDirPathStr
    ) -> DirPath:
        return DirPath(
            path=path, file_system=self.file_system
        )

    def build_like(
            self,
            path: t.Union[str, _path_strs.FileOrDirAbsPathT]
    ) -> t.Union[FilePath, DirPath]:
        path = _path_strs.FileOrDirAbsPathTypeAdapter.validate_python(path)
        if isinstance(path, _path_strs.AbsoluteDirPathStr):
            return self.build_dir_like(path=path)
        elif isinstance(path, _path_strs.AbsoluteFilePathStr):
            return self.build_file_like(path=path)
        else:
            raise ValueError(
                f"unexpected path type. {type(path)}, path: '{path}'"
            )

    def get_parent(self) -> DirPath:
        parent_path = self.path.get_parent()
        return self.build_dir_like(path=parent_path)

    @abstractmethod
    def get_name(self) -> str:
        """get simple name, i.e. the most terminal chunk in the path"""
        ...

    def ensure_exists(self) -> None:
        self.file_system.ensure_exists(self.path)

    def as_str(self) -> str:
        return self.file_system.contextualize_abs_path(self.path)

    def __get_logger_data__(self) -> t.Dict:
        return {
            "type": self.__class__.__name__,  # noqa
            "path": self.as_str(),
            "file_system_info": self.file_system.__get_logger_data__()
        }

    @classmethod
    @abstractmethod
    def init_from_str(cls, raw_str: str) -> _AbstractFileSystemObj:
        ...


class FilePath(_AbstractFileSystemObj):
    """
    this class represents a file on a file system,
    with the absolute full path string, and the file system instance
    """
    path: _path_strs.AbsoluteFilePathStr

    def _special_repr_short(self) -> str:
        return f"File[path='{self.path}', file_system_type={self.file_system.special_repr()}]"

    def _special_repr_verbose(self) -> str:
        return f"File[path='{self.path}', file_system_type={self.file_system.special_repr()}]"

    def delete(
            self,
            if_non_existent: t.Literal["raise", "return"] = "return",
    ) -> None:
        try:
            return self.file_system.delete_file(
                path=self.path,
                if_non_existent=if_non_existent
            )
        except Exception as e:
            raise e
        #
        # except _exceptions.FileSystemException as e:
        #     raise e
        # except Exception as e:
        #     raise _exceptions.FileSystemException(
        #         file_system=self.file_system,
        #         extra_info={
        #             "path": self.path,
        #             "error_type": type(e).__name__,
        #             "error_detail": str(e)
        #         }
        #     )

    def open(
            self,
            mode: _open_modes.SupportedOpenModesT = "rb"
    ) -> _open_modes.OpenFileT:
        try:
            return self.file_system.open_file(
                path=self.path,
                mode=mode
            )
        except Exception as e:
            raise e
        #
        # except _exceptions.FileSystemException as e:
        #     raise e
        # except Exception as e:
        #     raise _exceptions.FileSystemException(
        #         file_system=self.file_system,
        #         extra_info={
        #             "path": self.path,
        #             "error_type": type(e).__name__,
        #             "error_detail": str(e)
        #         }
        #     )

    def write(self, data: t.Union[str, bytes], mode: t.Literal["wb", "ab"] = "wb") -> None:
        byte_data: bytes
        if isinstance(data, bytes):
            byte_data = data
        else:
            byte_data = data.encode()

        with self.open(mode) as f:
            f.write(byte_data)

    def read_as_str(self, encoding: str = "utf-8") -> str:
        with self.open("rb") as f:
            byte_data = f.read()
        return byte_data.decode(encoding)

    def get_name(self, include_suffixes: bool = False) -> str:
        return self.path.get_name(include_suffixes=include_suffixes)

    def download_locally(
            self,
            dest_dir: _path_strs.AbsoluteDirPathStr
    ) -> _path_strs.AbsoluteFilePathStr:
        return self.file_system.download_file_locally(
            path=self.path,
            dest_dir=dest_dir
        )

    def ensure_correct_file_ext(self, choices: list[str]) -> None:
        choices = [
            _file_info.FileExt(choice) for choice in choices
        ]
        self.file_system.ensure_correct_file_ext(self.path, choices=choices)

    def ensure_correct_mime_type(self, choices: list[str]) -> None:
        choices = [
            _file_info.MimeType(choice) for choice in choices
        ]
        self.file_system.ensure_correct_mime_type(self.path, choices=choices)

    def get_file_ext(self) -> t.Optional[_file_info.FileExt]:
        return self.path.get_file_ext()

    def get_mime_type(self) -> t.Optional[_file_info.MimeType]:
        return self.path.get_mime_type()

    @classmethod
    def init_from_str(cls, raw_str: str) -> FilePath:
        file_path = _path_strs.AbsoluteFilePathStr(raw_str)
        return FilePath(path=file_path)


class DirPath(_AbstractFileSystemObj):
    """
    this class represents a directory on a file system,
    with the absolute full path string, and the file system instance
    """
    path: _path_strs.AbsoluteDirPathStr

    def _special_repr_short(self) -> str:
        return f"Dir[path='{self.path}', file_system_type={self.file_system.special_repr()}]"

    def _special_repr_verbose(self) -> str:
        return f"Dir[path='{self.path}', file_system_type={self.file_system.special_repr()}]"

    def delete(
            self,
            if_non_existent: t.Literal["raise", "return"] = "return",
            recursive: bool = True,
    ) -> None:
        try:
            return self.file_system.delete_dir(
                path=self.path,
                recursive=recursive,
                if_non_existent=if_non_existent
            )
        except Exception as e:
            raise e
        #
        # except _exceptions.FileSystemException as e:
        #     raise e
        # except Exception as e:
        #     raise _exceptions.FileSystemException(
        #         file_system=self.file_system,
        #         extra_info={
        #             "path": self.path,
        #             "error_type": type(e).__name__,
        #             "error_detail": str(e)
        #         }
        #     )

    def iter_dir(self) -> t.Iterator[t.Union[FilePath, DirPath]]:
        for path_i in self.file_system.iter_dir_contents(path=self.path):
            yield self.build_like(path=path_i)

    def iter_dir_contents_files_only(
            self,
            recursive: bool = True
    ) -> list[FilePath]:
        for path_i in self.file_system.iter_dir_contents_files_only(
            path=self.path,
            recursive=recursive
        ):
            yield self.build_like(path=path_i)

    @t.overload
    def joinpath(self, relative_path: _path_strs.RelFilePathStr) -> FilePath:
        ...

    @t.overload
    def joinpath(self, relative_path: _path_strs.RelDirPathStr) -> DirPath:
        ...

    @t.overload
    def joinpath(self, relative_path: str) -> t.Union[FilePath, DirPath]:
        ...

    def joinpath(
            self,
            relative_path: t.Union[str, _path_strs.FileOrDirRelPathT]
    ) -> t.Union[FilePath, DirPath]:
        rel_path_obj = _path_strs.FileOrDirRelPathTypeAdapter.validate_python(
            relative_path
        )
        new_path_any = self.path.joinpath(rel_path_obj)
        if isinstance(new_path_any, _path_strs.AbsoluteDirPathStr):
            new_path = new_path_any
            return DirPath(
                path=new_path,
                file_system=self.file_system
            )
        elif isinstance(new_path_any, _path_strs.AbsoluteFilePathStr):
            new_path = new_path_any

            return FilePath(
                path=new_path,
                file_system=self.file_system
            )
        else:
            raise ValueError(
                f"'{new_path_any}' is not valid type, type: {type(new_path_any)}"
            )

    def get_name(self) -> str:
        """get simple name, i.e. the most terminal chunk in the path"""
        return self.path.get_name()

    def make_self(
            self,
            if_exists: t.Literal["raise", "return"] = "return",
    ) -> None:
        try:
            self.file_system.make_dir(path=self.path, if_exists=if_exists)
        except Exception as e:
            raise e
        #
        # except _exceptions.FileSystemException as e:
        #     raise e
        # except Exception as e:
        #     raise _exceptions.FileSystemException(
        #         file_system=self.file_system,
        #         extra_info={
        #             "path": self.path,
        #             "error_type": type(e).__name__,
        #             "error_detail": str(e)
        #         }
        #     )

    def download_locally(
            self,
            dest_dir: _path_strs.AbsoluteDirPathStr
    ) -> _path_strs.AbsoluteDirPathStr:
        return self.file_system.download_dir_locally(
            path=self.path,
            dest_dir=dest_dir
        )

    @classmethod
    def init_from_str(cls, raw_str: str) -> DirPath:
        dir_path = _path_strs.AbsoluteDirPathStr(raw_str)
        return DirPath(path=dir_path)


def _str_parser(data: t.Any) -> t.Any:
    if isinstance(data, str):
        path_str = _path_strs.FileOrDirAbsPathTypeAdapter.validate_python(data)
        if isinstance(path_str, _path_strs.AbsoluteFilePathStr):
            return FilePath(path=path_str)
        elif isinstance(path_str, _path_strs.AbsoluteDirPathStr):
            return DirPath(path=path_str)
        else:
            raise ValueError(f"'{path_str}' unexpected type: {type(path_str)}")

    else:
        return data


FileOrDirPathT = t.Annotated[
    _AbstractFileSystemObj.build_dispatched_ann(),
    pydantic.BeforeValidator(_str_parser)
]
"""extends standard dispatched type to support raw strings"""


FileOrDirPathTypeAdapter = pydantic.TypeAdapter(FileOrDirPathT)
"""gives us pydantic parsing logic outside a pydantic class"""



from __future__ import annotations
import uuid
import pydantic

__all__ = ("GuidMixin", )


class GuidMixin(pydantic.BaseModel):
    """
    adds a UUID4 field, 'guid' that is an uuid4 object
    """
    guid: pydantic.UUID4 = pydantic.Field(
        description="globally unique id field",
        default_factory=uuid.uuid4
    )



from __future__ import annotations
import typing as t
import pydantic
from loguru import logger, _logger  # noqa

from spice_rack import _bases
from spice_rack._logging import _log_aug, _log_level


__all__ = (
    "Logger",
)


_ExtraLogDataT = t.Optional[list[t.Any]]
_LogLevelOrStrT = t.Union[str, _log_level.LogLevel]
LOGGER_WRAPPER_DEPTH = 3


class Logger(_bases.value_model.ValueModelBase):
    _cached_service_name: t.ClassVar[str] = "unknown"
    service_name: str
    
    @property
    def logger_wrapper_depth(self) -> int:
        return 3
        
    @pydantic.model_validator(mode="before", )
    def _use_cached_service_name_if_not_specified(cls, data: t.Any) -> t.Any:
        if isinstance(data, dict):
            service_name: str
            existing_service_name_maybe = data.get("service_name")
            if not existing_service_name_maybe:
                service_name = cls._cached_service_name
            else:
                service_name = existing_service_name_maybe
            data["service_name"] = service_name
        return data

    def cache_service_name(self) -> None:
        """save service name as the cached class attribute"""
        type(self)._cached_service_name = self.service_name

    @property
    def loguru_logger(self) -> _logger.Logger:
        return logger

    @staticmethod
    def _format_log_extra_data(
            extra_data: _ExtraLogDataT
    ) -> _log_aug.ExtraLogDataContainer:
        res: list[_log_aug.ExtraLogData] = []
        if extra_data:
            for data_i in extra_data:
                formatted_aug = _log_aug.ExtraLogData.parse_from_raw_data(data_i)
                res.append(formatted_aug)
        return _log_aug.ExtraLogDataContainer.model_validate(res)

    def _call_loguru(
            self,
            level: _log_level.LogLevel,
            msg: str,
            depth: int,
            loguru_func_name: t.Literal["log", "exception"],
            extra_data: _ExtraLogDataT,
    ) -> None:
        # catch errors and logs them, without blowing
        try:
            extra_data = self._format_log_extra_data(extra_data).get_logger_repr()

        except Exception as e:
            self._log(
                level=_log_level.LogLevel.model_validate("error"),
                msg="failed to format the specified log augmentations",
                extra_data=[str(e)],
                # manually specify depth here to be 0 bc this is an error within this logger
                depth=0
            )
            return

        try:
            loguru_level: str = str(level).upper()
            loguru_logger_inst = self.loguru_logger.opt(depth=depth)

            # use bind based on: https://github.com/Delgan/loguru/issues/318
            loguru_logger_inst = loguru_logger_inst.bind(service_name=self.service_name)
            if loguru_func_name == "log":
                loguru_logger_inst.log(
                    loguru_level, msg,
                    service_name=self.service_name,
                    extra_data=extra_data
                )
            elif loguru_func_name == "exception":
                loguru_logger_inst.exception(
                    msg, service_name=self.service_name, extra_data=extra_data
                )
            else:

                raise ValueError(
                    f"loguru_func_name must be 'log' or 'exception', "
                    f"encountered '{loguru_func_name}'"
                )

        except Exception as e:
            self._log(
                level=_log_level.LogLevel.model_validate("error"),
                msg="error encountered when calling the loguru logger",
                extra_data=[str(e)],

                # manually specify depth here to be 0 bc this is an error within this logger
                depth=0
            )
            return

    def _log(
            self,
            level: _log_level.LogLevel,
            msg: str,
            extra_data: _ExtraLogDataT,
            depth: t.Optional[int] = None,
    ) -> None:
        if depth is None:
            depth = self.logger_wrapper_depth

        return self._call_loguru(
            level=level,
            msg=msg,
            extra_data=extra_data,
            depth=depth,
            loguru_func_name="log"
        )

    def log(
            self,
            level: _log_level.LogLevel,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ) -> None:
        return self._log(
            level=_log_level.LogLevel.model_validate(level),
            msg=msg,
            extra_data=extra_data,
            depth=depth
        )

    def debug(
            self,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ):
        self.log(
            level=_log_level.LogLevel("debug"),
            msg=msg,
            extra_data=extra_data,
            depth=depth
        )

    def info(
            self,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ):
        self.log(
            level=_log_level.LogLevel("info"),
            msg=msg,
            extra_data=extra_data,
            depth=depth
        )

    def error(
            self,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ):
        self.log(
            level=_log_level.LogLevel("error"),
            msg=msg,
            extra_data=extra_data,
            depth=depth
        )

    def critical(
            self,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ):
        self.log(
            level=_log_level.LogLevel("critical"),
            msg=msg,
            extra_data=extra_data,
            depth=depth
        )

    def exception(
            self,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ):
        return self._call_loguru(
            level=_log_level.LogLevel("error"),
            msg=msg,
            extra_data=extra_data,
            depth=depth,
            loguru_func_name="exception"
        )
